
/******************************************************************************\
*                                                                              *
*                            rk4fixed-cpp v.3.00 beta                          *
*                                                                              *
*     The C++ module for solving ordinary dif. equations and their systems     *
*                             by Runge-Kutta method                            *
*                                                                              *
*    Testing (beta) version on nov. 2024                                       *
*    'As-Is', OpenSource                                                       *
*                                                                              *
*    Andrey G. Sadovoy                                                         *
*    sadovoya@mtu-net.ru                                                       *
*    sadovoya.narod.ru                                                         *
*    sintreseng.narod.ru                                                       *
*                                                                              *
\******************************************************************************/


#ifndef RK4FIXED_HPP_INCLUDED
#define RK4FIXED_HPP_INCLUDED

#include<type_traits>
#include<vector>
#include<array>

namespace runge_kutta {

    template<typename T, std::size_t N>
    using Syst_t = void(*)(T t, const std::array<T,N>& X,
        std::array<T,N>& rightParts);

    template<typename T>
    using Matrix_t = std::vector<std::vector<T>>;

    template<typename T, std::size_t N> //N is a count of equations
    int rk4fixed(Syst_t<T,N> syst,      //system to solve (fun. ptr)
        const std::array<T,N>& initConds, //initial conditions
        T t_init, T t_last,    //initial and last interval points
        std::size_t pointsCnt, //count of points for output
        std::vector<T>& tOut, Matrix_t<T>& XsOut, //results
        std::size_t stepsFact = 1)  //factor for steps inside of method
    {
        constexpr std::size_t one{1};

        static_assert(std::is_floating_point<T>::value, "T must be real");
        static_assert(N >= one, "N must be one at least");

        if( (pointsCnt <= one) || (!stepsFact) )
            return 1000;
            //Error Code 1000: Bad data.

        if( t_last < t_init )
            return 2000;
            //Error Code 2000: Last point must be not less init. point.

        tOut.resize(pointsCnt);
        XsOut.resize(N);
        for(auto& items : XsOut)
            items.resize(pointsCnt);

        std::array<T,N> varX, tmpX,
            c1, c2, c3, c4;

        const auto delt{(t_last-t_init)/T((pointsCnt-1)*stepsFact)};//inner step
        const auto halfDelt{T(.5)*delt};
        const auto deltBy6{delt/T(6.)};
        constexpr std::size_t null{0};
        std::size_t i, j, i_out;

        //Initial values of vars and first point of results:
        auto t = tOut[null] = t_init;
        for(j = null; j < N; ++j)
            XsOut[j][null] = varX[j] = initConds[j];

        for(i_out = one; i_out < pointsCnt; ++i_out) {
            /*t = t_init +(i_out-1)*delt*stepsFact; //else var*/
            for(i = null; i < stepsFact; ++i) {
                syst(t,varX,c1);
                for(j = null; j < N; ++j)
                    tmpX[j] = varX[j]+c1[j]*halfDelt;
                t += halfDelt;
                syst(t,tmpX,c2);
                for(j = null; j < N; ++j)
                    tmpX[j] = varX[j]+c2[j]*halfDelt;
                syst(t,tmpX,c3);
                for(j = null; j < N; ++j)
                    tmpX[j] = varX[j]+c3[j]*delt;
                t += halfDelt;
                syst(t,tmpX,c4);
                for(j = null; j < N; ++j) {
                    const auto c23{c2[j]+c3[j]};
                    varX[j] += (c1[j]+c23+c23+c4[j])*deltBy6;
                }
            }
            tOut[i_out] = t;
            for(j = null; j < N; ++j)
                XsOut[j][i_out] = varX[j];
        }

        return 0; //Error Code 0: No errors.
    }

}//namespace runge_kutta

#endif // RK4FIXED_HPP_INCLUDED
