//C++14 or higher needed
#include "../rk4fixed.hpp"
#include "helper.h"
#include <iostream>
#include <chrono>


namespace //for this file
{

    using FLOAT_TYPE = double; //you can try 'long double',for example, instead.

    using namespace runge_kutta;

    using std::cout;
    using std::cin;
    using std::endl;
    using std::size_t;
    using std::array;

    constexpr size_t MLN{1'000'000}; //C++14 or higher needed for ' separ.
    //can be changed to 1000000 if not supported

    //======================First problem for solving ==========================

    namespace problem_1
    {

        constexpr size_t num{2}; //count of eq. in sys.

        //system for solving
        void syst(FLOAT_TYPE t, const array<FLOAT_TYPE,num>& X,
            array<FLOAT_TYPE,num>& rp)
        {
            rp[0] = -1.0/(t*t);      //means dX0/dt = -1/t^2
            rp[1] = -2.0*X[0]/(t*t); //means dX1/dt = -2X0/t^2
        }

        //init. cond. for system
        constexpr array<FLOAT_TYPE,num> iC{1., 1.};

        //solution must be: X0(t) = 1/t, X1(t) = 1/t^2

        //beg. & end of calc. range
        constexpr FLOAT_TYPE t0{1.}, te{11.};

    }//namespace problem_1 end

    //======================Second problem for solving =========================

    namespace problem_2
    {
        constexpr size_t num{4}; //count of eq. in sys.

        //system for solving
        void syst(FLOAT_TYPE t, const array<FLOAT_TYPE,num>& X,
            array<FLOAT_TYPE,num>& rp)
        {
            rp[0] = 2.0*t;
            //means dX0/dt = 2t
            rp[1] = 3.0*(X[0]+t*t)/2.0;
            //means dX1/dt = 3(X0 + t^2)/2
            rp[2] = 4.0*(X[1]+t*X[0]+t*t*t)/3.0;
            //means dX2/dt = 4(X1 + t*X0 + t^3)/3
            rp[3] = 5.0*(X[2]+t*X[1]+t*t*X[0]+t*t*t*t)/4.0;
            //means dX3/dt = 5(X2 + t*X1 + t^2*X0 + t^4)/4
        }

        //init. cond. for syst.
        constexpr array<FLOAT_TYPE,num> iC{1., 1., 1., 1.};

        //Solution must be: X0 = t^2, X1 = t^3, X2 = t^4, X3 = t^5

        //beg. & end of calc. range
        constexpr FLOAT_TYPE t0{1.}, te{11.};

    }//namespace problem_2 end


    //for prints results
    void print_results_table(const std::vector<FLOAT_TYPE>& t,
        const Matrix_t<FLOAT_TYPE>& X, std::streamsize prec = 15)
    {
        size_t i,j;
        cout << "Table with columns by t, ";
        for(i = 0; i < X.size() - 1; ++i)
            cout << "X" << i << ", ";
        cout << "X" << X.size()-1 << " :\n\n";
        const auto old_prec{cout.precision(prec)};
        cout << std::showpoint;
        for(i = 0; i < t.size(); ++i)
        {
            cout << t[i];
            for(j = 0; j < X.size(); ++j)
                cout << '\t' << X[j][i];
            cout << endl;
        }
        (void)cout.precision(old_prec);
        cout << std::noshowpoint;
    }

    //solves the problem and displays the results
    template<size_t N>
    void solver(Syst_t<FLOAT_TYPE,N> system,
        const array<FLOAT_TYPE,N>& initConds, FLOAT_TYPE t_init,
            FLOAT_TYPE t_last, size_t pointsCnt, size_t stepsFact = 1,
                std::streamsize prec = 15)
    {
        using std::chrono::high_resolution_clock;
        using std::chrono::duration_cast;
        using std::chrono::duration;
        using std::chrono::nanoseconds;
        //using rep_ms = std::chrono::nanoseconds::rep;

        std::vector<FLOAT_TYPE> tOut;
        Matrix_t<FLOAT_TYPE> XsOut;

        const auto t0{high_resolution_clock::now()};

        const int erc = rk4fixed(system, initConds, t_init,
            t_last, pointsCnt, tOut, XsOut, stepsFact);
        //rk4fixed<FLOAT_TYPE,N> - FLOAT_TYPE,N deduced

        const auto time_ns =
            duration_cast<nanoseconds>(high_resolution_clock::now()-t0).count();

        if( !erc )
        {
            cout << "All steps inside method : " << (pointsCnt-1)*stepsFact
                << "\nTime : " << time_ns*1E-9 /*to seconds*/ << " seconds"
                    << "\nReturned with no errors\n\n";
            print_results_table(tOut, XsOut, prec);
        }
        else
            cout << "Returned with error (error code is " << erc << ")" << endl;
    }

} //namespace end



int main() {

    try
    {
        //number of points for output
        constexpr size_t pointsCnt_default{11};

        //factor for steps for calc. inside rk4fixed<>()
        constexpr size_t
            stepsFact_default{MLN}; //!!!deliberately greatly overstated!!!
                                    //(10 is good choice)

        size_t pointsCnt{pointsCnt_default}, stepsFact{stepsFact_default};

        for(;;)
        {
            cout << "Solution of Problem 1\n";
            { //Solving of first problem:
                using namespace problem_1;
                solver(&syst, iC, t0, te, pointsCnt, stepsFact);
                //solver<num> - num deduced
            }

            cout << "\n\nSolution of Problem 2\n";

            { //Solving of 2-nd problem:
                using namespace problem_2;
                solver(&syst, iC, t0, te, pointsCnt, stepsFact);
                //solver<num> - num deduced
            }

            if( !request_repeate() )
            {
                pointsCnt = request_cin<size_t>("Enter number of points for "
                    "output [0 for exit]: ");
                if( !pointsCnt )
                    break;
                stepsFact = request_cin<size_t>("Enter factor for steps for "
                    "calculations: ");
            }

            cout << endl;

        }

    }
    catch(...)
    { //try-catch for stack unwinding on exception
        return -1000;
    }

    return 0;
}

